from typing import Iterable, Literal, Optional

Endianness = Literal["big", "little"]
HASH_BYTES = 32


def to_bytes(
    value: int,
    length: Optional[int] = None,
    byte_order: Optional[Endianness] = None,
    signed: Optional[bool] = None,
) -> bytes:
    """
    Converts the given integer to a bytes object of given length and byte order.
    The default values are 32B width (which is the hash result width) and 'big', respectively.
    """
    if length is None:
        length = HASH_BYTES

    if byte_order is None:
        byte_order = "big"

    if signed is None:
        signed = False

    return int.to_bytes(value, length=length, byteorder=byte_order, signed=signed)


def from_bytes(
    value: bytes,
    byte_order: Optional[Endianness] = None,
    signed: Optional[bool] = None,
) -> int:
    """
    Converts the given bytes object (parsed according to the given byte order) to an integer.
    Default byte order is 'big'.
    """
    if byte_order is None:
        byte_order = "big"

    if signed is None:
        signed = False

    return int.from_bytes(value, byteorder=byte_order, signed=signed)


def blockify(data, chunk_size: int) -> Iterable:
    """
    Returns the given data partitioned to chunks of chunks_size (last chunk might be smaller).
    """
    assert chunk_size > 0, f"chunk_size must be greater than 0. Got: {chunk_size}."
    return (data[i : i + chunk_size] for i in range(0, len(data), chunk_size))



from typing import List
from poseidon_py.c_bindings import hades_permutation
from poseidon_py.utils import blockify, from_bytes, to_bytes

POSEIDON_PARAMS = {
    "m": 3,
    "r": 2,
}


def poseidon_perm(x: int, y: int, z: int) -> List[int]:
    """
    Returns the poseidon permutation of the inputs.
    """
    return hades_permutation([x, y, z])


def poseidon_hash_func(x: bytes, y: bytes) -> bytes:
    """
    Returns the poseidon_hash of the inputs.
    """
    return to_bytes(poseidon_perm(from_bytes(x), from_bytes(y), 2)[0])


def poseidon_hash(x: int, y: int) -> int:
    """
    Hashes two elements and retrieves a single field element output.
    """
    return hades_permutation([x, y, 2])[0]


def poseidon_hash_single(x: int) -> int:
    """
    Hashes single element and retrieves a single field element output.
    """
    return hades_permutation([x, 0, 1])[0]


def poseidon_hash_many(array: List[int]) -> int:
    """
    Hashes array of elements and retrieves a single field element output.
    """
    values = list(array)
    m = POSEIDON_PARAMS["m"]
    r = POSEIDON_PARAMS["r"]

    # Pad input with 1 followed by 0's (if necessary).
    values.append(1)
    values += [0] * (-len(values) % r)

    assert len(values) % r == 0
    state = [0] * m
    for block in blockify(data=values, chunk_size=r):
        state = list(
            hades_permutation(
                [state_val + block_val for state_val, block_val in zip(state, block)]
                + state[-1:]
            )
        )

    return state[0]



import subprocess
import setuptools
from setuptools.command.build_py import build_py
from setuptools.command.build_ext import build_ext

class PoseidonExtension(setuptools.Extension):
    def __init__(self):
        self.name = "poseidon"
        setuptools.Extension.__init__(self, self.name, sources=[])


class BuildPy(build_py):
    def run(self):
        self.run_command("build_ext")
        return super().run()


class BuildPoseidon(build_ext):
    already_built = False

    def build_extension(self, ext):
        if self.already_built:
            return

        subprocess.run("chmod a+x ./build.sh && ./build.sh", shell=True, check=True)
        self.already_built = True


if __name__ == "__main__":
    with open("README.rst", encoding="utf-8") as f:
        long_description = f.read()

    setuptools.setup(
        name="poseidon_py",
        version="0.1.1",
        description="Python implementation of Poseidon hash",
        long_description=long_description,
        author="drknzz",
        author_email="kamil.jankowski.x@gmail.com",
        url="https://github.com/drknzz/poseidon-py.git",
        extras_require={"build": ["make"]},
        ext_modules=[PoseidonExtension()],
        cmdclass={
            "build_py": BuildPy,
            "build_ext": BuildPoseidon,
        },
        python_requires=">=3.9",
        packages=["poseidon_py"],
        package_data={"poseidon_py": ["../lib_pos.*"]},
    )



import ctypes
import dataclasses
import pathlib
import struct
import os
from typing import List, Optional

LIB_NAME = "lib_pos"
UINT256_MAX = 2**256 - 1


def hades_permutation(values: List[int]) -> List[int]:
    _validate_values(values)

    LOADER.load_c_lib()

    c_values = Converter.make_c_values(values)

    assert LOADER.c_lib is not None
    LOADER.c_lib.permutation_3(c_values)

    return Converter.make_py_values(c_values)


def _validate_values(values: List[int]) -> None:
    if len(values) != 3:
        raise ValueError(f"expected 3 values, got {len(values)} values")

    for val in values:
        _assert_u_int256_t(val)


def _assert_u_int256_t(value: int) -> None:
    if not 0 <= value <= UINT256_MAX:
        raise ValueError(f"{value} out of u_int256_t range")


@dataclasses.dataclass
class Loader:
    lib_name: str
    c_lib: Optional[ctypes.CDLL] = None

    def load_c_lib(self) -> None:
        if self.c_lib:
            return

        lib_path = self._get_lib_path()
        self.c_lib = ctypes.CDLL(str(lib_path))

        self._set_types()

    def _get_lib_path(self) -> pathlib.Path:
        dir_ = pathlib.Path(__file__).parents[1]
        file_ = next(file for file in os.listdir(dir_) if file.startswith(LIB_NAME))
        return dir_ / file_

    def _set_types(self) -> None:
        assert self.c_lib is not None
        self.c_lib.permutation_3.argtypes = [ctypes.POINTER(Converter.felt_t)]
        self.c_lib.permutation_3.restype = None


LOADER = Loader(LIB_NAME)


class Converter:
    felt_t = ctypes.c_uint64 * 4

    @staticmethod
    def make_c_values(values: List[int]) -> ctypes.Array[felt_t]:
        felts = [Converter.int_to_felt_t(val) for val in values]
        return (Converter.felt_t * 3)(*felts)

    @staticmethod
    def int_to_felt_t(value: int) -> felt_t:
        value_bytes = value.to_bytes(32, byteorder="little", signed=False)
        felt_ = struct.unpack("4Q", value_bytes)
        return Converter.felt_t(*felt_)

    @staticmethod
    def make_py_values(c_values: ctypes.Array[felt_t]) -> List[int]:
        return [
            int.from_bytes(val, byteorder="little", signed=False) for val in c_values
        ]
